@@ -20,6 +20,7 @@ from fhirpath.fhirspec import (
     FHIRSearchSpecFactory,
     ResourceSearchParameterDefinition,
     SearchParameter,
+    search_param_prefixes,
 )
 from fhirpath.fql import (
     G_,
@@ -162,7 +163,7 @@ class SearchContext(object):
                 raw_value = raw_value[0]
 
             values: List = list()
-            self.normalize_param_value(raw_value, values)
+            self.normalize_param_value(raw_value, sp, values)
 
             if len(values) == 1:
                 param_value_ = values[0]
@@ -174,12 +175,14 @@ class SearchContext(object):
             normalized_params.append((_path, param_value_, modifier_))
         return normalized_params
 
-    def normalize_param_value(self, raw_value: Union[List, str], container):
+    def normalize_param_value(
+        self, raw_value: Union[List, str], search_param: SearchParameter, container
+    ):
         """ """
         if isinstance(raw_value, list):
             bucket: List[str] = list()
             for rv in raw_value:
-                self.normalize_param_value(rv, bucket)
+                self.normalize_param_value(rv, search_param, bucket)
             if len(bucket) == 1:
                 container.append(bucket[0])
             else:
@@ -201,8 +204,8 @@ class SearchContext(object):
                 else:
                     val_ = val
 
-                for prefix in value_prefixes:
-                    if val_.startswith(prefix):
+                for prefix in search_param_prefixes:
+                    if val_.startswith(prefix) and search_param.support_prefix():
                         comparison_operator = prefix
                         val_ = val_[2:]
                         break
@@ -250,7 +253,7 @@ class SearchContext(object):
             value_parts[0],
         ]
         part1_param_value = list()
-        self.normalize_param_value(part1[1], part1_param_value)
+        self.normalize_param_value(part1[1], param_def, part1_param_value)
         if len(part1_param_value) == 1:
             part1_param_value = part1_param_value[0]
         composite_bucket.append(
@@ -264,7 +267,7 @@ class SearchContext(object):
             ]
             part2.append(part_)
         part2_param_value = list()
-        self.normalize_param_value(part2[0][1], part2_param_value)
+        self.normalize_param_value(part2[0][1], param_def, part2_param_value)
 
         if len(part2_param_value) == 1:
             part2_param_value = part2_param_value[0]
@@ -658,6 +661,10 @@ class Search(object):
             # Extract IDs from the main query result
             ids = main_query_result.extract_ids()
 
+            # if no IDs were extracted from the main_query_result, skip.
+            if not ids:
+                continue
+
             # Build a Q_ (query) object to join the resource based on reference ids.
             builder = Q_([from_resource_type], self.context.engine)
             terms: List = []
@@ -729,6 +736,8 @@ class Search(object):
                 term_factory = self.create_humanname_term
             elif klass_name == "Money":
                 term_factory = self.create_money_term
+            elif klass_name == "Period":
+                term_factory = self.create_period_term
             else:
                 raise NotImplementedError(
                     f"Can't perform search on element of type {klass_name}"
@@ -905,9 +914,7 @@ class Search(object):
 
         raise NotImplementedError
 
-    def single_valued_coding_term(
-        self, path_, value, modifier, ignore_not_modifier=False
-    ):
+    def single_valued_coding_term(self, path_, value, modifier):
         """ """
         operator_, original_value = value
 
@@ -920,6 +927,7 @@ class Search(object):
 
         has_pipe = "|" in original_value
         terms = list()
+        subpredicate_modifier = None if modifier == "not" else modifier
 
         if modifier == "text" and not has_pipe:
             # xxx: should be validation error if value contained pipe
@@ -934,14 +942,14 @@ class Search(object):
                 path_1 = path_ / "code"
                 new_value = (value[0], original_value[1:])
 
-                term = self.create_term(path_1, new_value, modifier)
+                term = self.create_term(path_1, new_value, subpredicate_modifier)
                 terms.append(term)
 
             elif original_value.endswith("|"):
                 path_1 = path_ / "system"
                 new_value = (value[0], original_value[:-1])
 
-                terms.append(self.create_term(path_1, new_value, modifier))
+                terms.append(self.create_term(path_1, new_value, subpredicate_modifier))
 
             else:
                 parts = original_value.split("|")
@@ -949,22 +957,22 @@ class Search(object):
                 try:
                     path_1 = path_ / "system"
                     new_value = (value[0], parts[0])
-                    term = self.create_term(path_1, new_value, modifier)
+                    term = self.create_term(path_1, new_value, subpredicate_modifier)
                     terms.append(term)
 
                     path_2 = path_ / "code"
                     new_value = (value[0], parts[1])
-                    term = self.create_term(path_2, new_value, modifier)
+                    term = self.create_term(path_2, new_value, subpredicate_modifier)
                     terms.append(term)
 
                 except IndexError:
                     pass
         else:
             path_1 = path_ / "code"
-            terms.append(self.create_term(path_1, value, modifier))
+            terms.append(self.create_term(path_1, value, subpredicate_modifier))
 
         group = G_(*terms, path=path_, type_=GroupType.COUPLED)
-        if modifier == "not" and ignore_not_modifier is False:
+        if modifier == "not":
             group.match_operator = MatchType.NONE
         return group
 
@@ -1278,6 +1286,93 @@ class Search(object):
         assert self.context.engine.fhir_release == FHIR_VERSION.STU3
         return self.single_valued_quantity_term(path_, value, modifier)
 
+    def create_period_term(self, path_, param_value, modifier):
+        if isinstance(param_value, list):
+            terms = [
+                self.single_valued_period_term(path_, value, modifier)
+                for value in param_value
+            ]
+            return G_(*terms, path=path_, type_=GroupType.COUPLED)
+
+        elif isinstance(param_value, tuple):
+            return self.single_valued_period_term(path_, param_value, modifier)
+
+        raise NotImplementedError
+
+    def single_valued_period_term(self, path_, value, modifier):
+        operator, original_value = value
+
+        if isinstance(original_value, list):
+            terms = [
+                self.single_valued_period_term(path_, val, modifier)
+                for val in original_value
+            ]
+            # IN Like Group
+            return G_(*terms, path=path_, type_=GroupType.DECOUPLED)
+
+        if operator == "eq":
+            terms = [
+                self.create_term(path_ / "start", ("ge", original_value), modifier),
+                self.create_term(path_ / "end", ("le", original_value), modifier),
+            ]
+            type_ = GroupType.COUPLED
+        elif operator == "ne":
+            terms = [
+                self.create_term(path_ / "start", ("lt", original_value), modifier),
+                self.create_term(path_ / "end", ("gt", original_value), modifier),
+                not_exists_(path_ / "start"),
+                not_exists_(path_ / "end"),
+            ]
+            type_ = GroupType.DECOUPLED
+        elif operator == "gt":
+            terms = [
+                self.create_term(path_ / "end", ("gt", original_value), modifier),
+                not_exists_(path_ / "end"),
+            ]
+            type_ = GroupType.DECOUPLED
+        elif operator == "lt":
+            terms = [
+                self.create_term(path_ / "start", ("lt", original_value), modifier),
+            ]
+            type_ = GroupType.COUPLED
+        elif operator == "ge":
+            terms = [
+                self.create_term(path_ / "end", ("ge", original_value), modifier),
+                not_exists_(path_ / "end"),
+            ]
+            type_ = GroupType.DECOUPLED
+        elif operator == "le":
+            terms = [
+                self.create_term(path_ / "start", ("le", original_value), modifier),
+            ]
+            type_ = GroupType.COUPLED
+        elif operator == "sa":
+            terms = [
+                self.create_term(path_ / "start", ("gt", original_value), modifier),
+            ]
+            type_ = GroupType.COUPLED
+        elif operator == "eb":
+            terms = [
+                self.create_term(path_ / "end", ("lt", original_value), modifier),
+            ]
+            type_ = GroupType.COUPLED
+        elif operator == "ap":
+            start_terms = [
+                self.create_term(path_ / "start", ("le", original_value), modifier)
+            ]
+            start_group = G_(*start_terms, path=path_, type_=GroupType.COUPLED)
+            end_terms = [
+                self.create_term(path_ / "end", ("ge", original_value), modifier),
+                not_exists_(path_ / "end"),
+            ]
+            end_group = G_(*end_terms, path=path_, type_=GroupType.DECOUPLED)
+            terms = [start_group, end_group]
+            type_ = GroupType.COUPLED
+        else:
+            raise NotImplementedError(f"prefix {operator} not handled for periods.")
+
+        return G_(*terms, path=path_, type_=type_)
+
     def validate_pre_term(self, operator_, path_, value, modifier):
         """ """
         if modifier in ("above", "below") and operator_ in ("sa", "eb"):
@@ -1305,7 +1400,8 @@ class Search(object):
                 # xxx: should be validated already
                 terms_ = list()
                 for val in original_value:
-                    term_ = self.create_term(path_, val, modifier)
+                    subpredicate_modifier = None if modifier == "not" else modifier
+                    term_ = self.create_term(path_, val, subpredicate_modifier)
                     terms_.append(term_)
                 # IN Like Group
                 group = G_(*terms_, path=path_, type_=GroupType.DECOUPLED)
@@ -1437,11 +1533,13 @@ class Search(object):
                 offset = (current_page - 1) * self.result_params["_count"]
         return builder.limit(self.result_params["_count"], offset)
 
-    def response(self, result, includes):
+    def response(self, result, includes, as_json):
         """ """
-        return self.context.engine.wrapped_with_bundle(result, includes)
+        return self.context.engine.wrapped_with_bundle(
+            result, includes=includes, as_json=as_json
+        )
 
-    def __call__(self):
+    def __call__(self, as_json=False):
         """ """
 
         # TODO: chaining
@@ -1467,7 +1565,9 @@ class Search(object):
             # but we should be more explicit about the query context.
             if not self.reverse_chaining_results:
                 return self.response(
-                    EngineResult(EngineResultHeader(total=0), EngineResultBody()), []
+                    EngineResult(EngineResultHeader(total=0), EngineResultBody()),
+                    [],
+                    as_json,
                 )
 
         # MAIN QUERY
@@ -1493,13 +1593,13 @@ class Search(object):
         ]
 
         all_includes = [*include_results, *rev_include_results]
-        return self.response(main_result, all_includes)
+        return self.response(main_result, all_includes, as_json)
 
 
 class AsyncSearch(Search):
     """ """
 
-    async def __call__(self):
+    async def __call__(self, as_json=False):
         """ """
         # TODO: chaining
 
@@ -1523,7 +1623,7 @@ class AsyncSearch(Search):
             # FIXME: we use the result of the last _has query to build the empty bundle,
             # but we should be more explicit about the query context.
             if not self.reverse_chaining_results:
-                return self.response(res, [])
+                return self.response(res, [], as_json)
 
         # MAIN QUERY
         self.main_query = self.build()
@@ -1548,7 +1648,7 @@ class AsyncSearch(Search):
         ]
 
         all_includes = [*include_results, *rev_include_results]
-        return self.response(main_result, all_includes)
+        return self.response(main_result, all_includes, as_json)
 
 
 def fhir_search(context, query_string=None, params=None):
