@@ -13,6 +13,7 @@ from types import ModuleType
 from typing import (
     TYPE_CHECKING,
     Any,
+    Callable,
     Dict,
     List,
     Match,
@@ -24,7 +25,9 @@ from typing import (
     cast,
 )
 
+import orjson
 import pkg_resources
+from pydantic.json import pydantic_encoder
 from pydantic.validators import bool_validator
 from yarl import URL
 from zope.interface import implementer
@@ -635,10 +638,57 @@ class BundleWrapper:
 
         return link
 
-    def __call__(self):
+    def __call__(self, as_json=False):
         """ """
+        if as_json:
+            # if as_json is True, return the bundle as python dict
+            # instead of building a pydantic.BaseModel
+            return self.data
         return self.bundle_model.parse_obj(self.data)
 
     def json(self):
         """ """
         return self.__call__().json()
+
+
+def json_dumps(
+    value: Union[List[Dict[str, Any]], Dict[str, Any]],
+    *,
+    default: Callable = None,
+    option: int = 0,
+    indent: int = None,
+    sort_keys: bool = None,
+    return_bytes: bool = False,
+) -> Union[str, bytes]:
+    """Special helper function to serialize to json str, powered by ``orjson``"""
+    if default is None:
+        default = pydantic_encoder
+
+    params: Any = {"default": default}
+    option_ = option
+    if option_ == 0:
+        if indent is not None:
+            # only indent 2 is accepted
+            option_ |= orjson.OPT_INDENT_2
+        if sort_keys is not None and sort_keys:
+            option_ |= orjson.OPT_SORT_KEYS
+
+    if option_ > 0:
+        params.update({"option": option_})
+
+    v = orjson.dumps(value, **params)
+    if return_bytes is False:
+        return force_str(v)
+    return v
+
+
+def json_loads(
+    value: Union[bytes, bytearray, str]
+) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
+    """bytes, bytearray, and str input are accepted.
+    If the input exists as bytes (was read directly from a source),
+    it is recommended to pass bytes. This has lower memory usage and lower latency.
+
+    The input must be valid UTF-8."""
+
+    return orjson.loads(value)
